@@ -33,16 +33,23 @@ License along with this library.  If not, see <http://www.gnu.org/licenses/>.
 #include "output.h"
 #include "plasmashell.h"
 #include "plasmawindowmanagement.h"
+#include "pointerconstraints.h"
+#include "pointergestures.h"
 #include "seat.h"
 #include "shadow.h"
 #include "blur.h"
 #include "contrast.h"
+#include "relativepointer.h"
 #include "server_decoration.h"
 #include "slide.h"
 #include "shell.h"
 #include "shm_pool.h"
 #include "subcompositor.h"
+#include "textinput_p.h"
+#include "xdgshell.h"
+#include "xdgshell_p.h"
 #include "wayland_pointer_p.h"
+#include "xdgforeign_v2.h"
 // Qt
 #include <QDebug>
 // wayland
@@ -60,6 +67,14 @@ License along with this library.  If not, see <http://www.gnu.org/licenses/>.
 #include <wayland-slide-client-protocol.h>
 #include <wayland-dpms-client-protocol.h>
 #include <wayland-server-decoration-client-protocol.h>
+#include <wayland-text-input-v0-client-protocol.h>
+#include <wayland-text-input-v2-client-protocol.h>
+#include <wayland-xdg-shell-v5-client-protocol.h>
+#include <wayland-xdg-shell-v6-client-protocol.h>
+#include <wayland-relativepointer-unstable-v1-client-protocol.h>
+#include <wayland-pointer-gestures-unstable-v1-client-protocol.h>
+#include <wayland-pointer-constraints-unstable-v1-client-protocol.h>
+#include <wayland-xdg-foreign-unstable-v2-client-protocol.h>
 
 /*****
  * How to add another interface:
@@ -138,14 +153,14 @@ static const QMap<Registry::Interface, SuppertedInterfaceData> s_interfaces = {
         &Registry::subCompositorRemoved
     }},
     {Registry::Interface::PlasmaShell, {
-        2,
+        4,
         QByteArrayLiteral("org_kde_plasma_shell"),
         &org_kde_plasma_shell_interface,
         &Registry::plasmaShellAnnounced,
         &Registry::plasmaShellRemoved
     }},
     {Registry::Interface::PlasmaWindowManagement, {
-        2,
+        7,
         QByteArrayLiteral("org_kde_plasma_window_management"),
         &org_kde_plasma_window_management_interface,
         &Registry::plasmaWindowManagementAnnounced,
@@ -159,7 +174,7 @@ static const QMap<Registry::Interface, SuppertedInterfaceData> s_interfaces = {
         &Registry::idleRemoved
     }},
     {Registry::Interface::FakeInput, {
-        1,
+        2,
         QByteArrayLiteral("org_kde_kwin_fake_input"),
         &org_kde_kwin_fake_input_interface,
         &Registry::fakeInputAnnounced,
@@ -180,7 +195,7 @@ static const QMap<Registry::Interface, SuppertedInterfaceData> s_interfaces = {
         &Registry::outputDeviceRemoved
     }},
     {Registry::Interface::Shadow, {
-        1,
+        2,
         QByteArrayLiteral("org_kde_kwin_shadow_manager"),
         &org_kde_kwin_shadow_manager_interface,
         &Registry::shadowAnnounced,
@@ -227,6 +242,69 @@ static const QMap<Registry::Interface, SuppertedInterfaceData> s_interfaces = {
         &org_kde_kwin_server_decoration_manager_interface,
         &Registry::serverSideDecorationManagerAnnounced,
         &Registry::serverSideDecorationManagerRemoved
+    }},
+    {Registry::Interface::TextInputManagerUnstableV0, {
+        1,
+        QByteArrayLiteral("wl_text_input_manager"),
+        &wl_text_input_manager_interface,
+        &Registry::textInputManagerUnstableV0Announced,
+        &Registry::textInputManagerUnstableV0Removed
+    }},
+    {Registry::Interface::TextInputManagerUnstableV2, {
+        1,
+        QByteArrayLiteral("zwp_text_input_manager_v2"),
+        &zwp_text_input_manager_v2_interface,
+        &Registry::textInputManagerUnstableV2Announced,
+        &Registry::textInputManagerUnstableV2Removed
+    }},
+    {Registry::Interface::XdgShellUnstableV5, {
+        1,
+        QByteArrayLiteral("xdg_shell"),
+        &xdg_shell_interface,
+        &Registry::xdgShellUnstableV5Announced,
+        &Registry::xdgShellUnstableV5Removed
+    }},
+    {Registry::Interface::RelativePointerManagerUnstableV1, {
+        1,
+        QByteArrayLiteral("zwp_relative_pointer_manager_v1"),
+        &zwp_relative_pointer_manager_v1_interface,
+        &Registry::relativePointerManagerUnstableV1Announced,
+        &Registry::relativePointerManagerUnstableV1Removed
+    }},
+    {Registry::Interface::PointerGesturesUnstableV1, {
+        1,
+        QByteArrayLiteral("zwp_pointer_gestures_v1"),
+        &zwp_pointer_gestures_v1_interface,
+        &Registry::pointerGesturesUnstableV1Announced,
+        &Registry::pointerGesturesUnstableV1Removed
+    }},
+    {Registry::Interface::PointerConstraintsUnstableV1, {
+        1,
+        QByteArrayLiteral("zwp_pointer_constraints_v1"),
+        &zwp_pointer_constraints_v1_interface,
+        &Registry::pointerConstraintsUnstableV1Announced,
+        &Registry::pointerConstraintsUnstableV1Removed
+    }},
+    {Registry::Interface::XdgExporterUnstableV2, {
+        1,
+        QByteArrayLiteral("zxdg_exporter_v2"),
+        &zxdg_exporter_v2_interface,
+        &Registry::exporterUnstableV2Announced,
+        &Registry::exporterUnstableV2Removed
+    }},
+    {Registry::Interface::XdgImporterUnstableV2, {
+        1,
+        QByteArrayLiteral("zxdg_importer_v2"),
+        &zxdg_importer_v2_interface,
+        &Registry::importerUnstableV2Announced,
+        &Registry::importerUnstableV2Removed
+    }},
+    {Registry::Interface::XdgShellUnstableV6, {
+        1,
+        QByteArrayLiteral("zxdg_shell_v6"),
+        &zxdg_shell_v6_interface,
+        &Registry::xdgShellUnstableV6Announced,
+        &Registry::xdgShellUnstableV6Removed
     }}
 };
 
@@ -240,7 +318,7 @@ static quint32 maxVersion(const Registry::Interface &interface)
 }
 }
 
-class Registry::Private
+class Q_DECL_HIDDEN Registry::Private
 {
 public:
     Private(Registry *q);
@@ -248,6 +326,7 @@ public:
     bool hasInterface(Interface interface) const;
     AnnouncedInterface interface(Interface interface) const;
     QVector<AnnouncedInterface> interfaces(Interface interface) const;
+    Interface interfaceForName(quint32 name) const;
     template <typename T>
     T *bind(Interface interface, uint32_t name, uint32_t version) const;
     template <class T, typename WL>
@@ -306,6 +385,7 @@ void Registry::release()
 
 void Registry::destroy()
 {
+    emit registryDestroyed();
     d->registry.destroy();
     d->callback.destroy();
 }
@@ -325,6 +405,7 @@ void Registry::create(wl_display *display)
 void Registry::create(ConnectionThread *connection)
 {
     create(connection->display());
+    connect(connection, &ConnectionThread::connectionDied, this, &Registry::destroy);
 }
 
 void Registry::setup()
@@ -468,6 +549,18 @@ Registry::AnnouncedInterface Registry::Private::interface(Interface interface) c
     return AnnouncedInterface{0, 0};
 }
 
+Registry::Interface Registry::Private::interfaceForName(quint32 name) const
+{
+    auto it = std::find_if(m_interfaces.constBegin(), m_interfaces.constEnd(),
+        [name] (const InterfaceData &data) {
+            return data.name == name;
+        });
+    if (it == m_interfaces.constEnd()) {
+        return Interface::Unknown;
+    }
+    return (*it).interface;
+}
+
 bool Registry::hasInterface(Registry::Interface interface) const
 {
     return d->hasInterface(interface);
@@ -506,6 +599,15 @@ BIND(FakeInput, org_kde_kwin_fake_input)
 BIND(OutputManagement, org_kde_kwin_outputmanagement)
 BIND(OutputDevice, org_kde_kwin_outputdevice)
 BIND(ServerSideDecorationManager, org_kde_kwin_server_decoration_manager)
+BIND(TextInputManagerUnstableV0, wl_text_input_manager)
+BIND(TextInputManagerUnstableV2, zwp_text_input_manager_v2)
+BIND(XdgShellUnstableV5, xdg_shell)
+BIND(XdgShellUnstableV6, zxdg_shell_v6)
+BIND(RelativePointerManagerUnstableV1, zwp_relative_pointer_manager_v1)
+BIND(PointerGesturesUnstableV1, zwp_pointer_gestures_v1)
+BIND(PointerConstraintsUnstableV1, zwp_pointer_constraints_v1)
+BIND(XdgExporterUnstableV2, zxdg_exporter_v2)
+BIND(XdgImporterUnstableV2, zxdg_importer_v2)
 BIND2(ShadowManager, Shadow, org_kde_kwin_shadow_manager)
 BIND2(BlurManager, Blur, org_kde_kwin_blur_manager)
 BIND2(ContrastManager, Contrast, org_kde_kwin_contrast_manager)
@@ -528,6 +630,7 @@ T *Registry::Private::create(quint32 name, quint32 version, QObject *parent, WL
             }
         }
     );
+    QObject::connect(q, &Registry::registryDestroyed, t, &T::destroy);
     return t;
 }
 
@@ -563,6 +666,72 @@ CREATE2(ShmPool, Shm)
 #undef CREATE
 #undef CREATE2
 
+XdgExporter *Registry::createXdgExporter(quint32 name, quint32 version, QObject *parent)
+{
+    //only V1 supported for now
+    return d->create<XdgExporterUnstableV2>(name, version, parent, &Registry::bindXdgExporterUnstableV2);
+}
+
+XdgImporter *Registry::createXdgImporter(quint32 name, quint32 version, QObject *parent)
+{
+    //only V1 supported for now
+    return d->create<XdgImporterUnstableV2>(name, version, parent, &Registry::bindXdgImporterUnstableV2);
+}
+
+TextInputManager *Registry::createTextInputManager(quint32 name, quint32 version, QObject *parent)
+{
+    switch (d->interfaceForName(name)) {
+    case Interface::TextInputManagerUnstableV0:
+        return d->create<TextInputManagerUnstableV0>(name, version, parent, &Registry::bindTextInputManagerUnstableV0);
+    case Interface::TextInputManagerUnstableV2:
+        return d->create<TextInputManagerUnstableV2>(name, version, parent, &Registry::bindTextInputManagerUnstableV2);
+    default:
+        return nullptr;
+    }
+}
+
+XdgShell *Registry::createXdgShell(quint32 name, quint32 version, QObject *parent)
+{
+    switch (d->interfaceForName(name)) {
+    case Interface::XdgShellUnstableV5:
+        return d->create<XdgShellUnstableV5>(name, version, parent, &Registry::bindXdgShellUnstableV5);
+    case Interface::XdgShellUnstableV6:
+        return d->create<XdgShellUnstableV6>(name, version, parent, &Registry::bindXdgShellUnstableV6);
+    default:
+        return nullptr;
+    }
+}
+
+RelativePointerManager *Registry::createRelativePointerManager(quint32 name, quint32 version, QObject *parent)
+{
+    switch (d->interfaceForName(name)) {
+    case Interface::RelativePointerManagerUnstableV1:
+        return d->create<RelativePointerManager>(name, version, parent, &Registry::bindRelativePointerManagerUnstableV1);
+    default:
+        return nullptr;
+    }
+}
+
+PointerGestures *Registry::createPointerGestures(quint32 name, quint32 version, QObject *parent)
+{
+    switch (d->interfaceForName(name)) {
+    case Interface::PointerGesturesUnstableV1:
+        return d->create<PointerGestures>(name, version, parent, &Registry::bindPointerGesturesUnstableV1);
+    default:
+        return nullptr;
+    }
+}
+
+PointerConstraints *Registry::createPointerConstraints(quint32 name, quint32 version, QObject *parent)
+{
+    switch (d->interfaceForName(name)) {
+    case Interface::PointerConstraintsUnstableV1:
+        return d->create<PointerConstraints>(name, version, parent, &Registry::bindPointerConstraintsUnstableV1);
+    default:
+        return nullptr;
+    }
+}
+
 namespace {
 static const wl_interface *wlInterface(Registry::Interface interface)
 {
