@@ -22,6 +22,7 @@ License along with this library.  If not, see <http://www.gnu.org/licenses/>.
 #include "buffer_interface.h"
 #include "clientconnection.h"
 #include "compositor_interface.h"
+#include "pointerconstraints_interface_p.h"
 #include "region_interface.h"
 #include "subcompositor_interface.h"
 #include "subsurface_interface_p.h"
@@ -174,6 +175,82 @@ void SurfaceInterface::Private::setContrast(const QPointer<ContrastInterface> &c
     pending.contrastIsSet = true;
 }
 
+void SurfaceInterface::Private::installPointerConstraint(LockedPointerInterface *lock)
+{
+    Q_ASSERT(lockedPointer.isNull());
+    Q_ASSERT(confinedPointer.isNull());
+    lockedPointer = QPointer<LockedPointerInterface>(lock);
+    if (lock->lifeTime() == LockedPointerInterface::LifeTime::OneShot) {
+        constrainsOneShotConnection = QObject::connect(lock, &LockedPointerInterface::lockedChanged, q_func(),
+            [this] {
+                if (lockedPointer.isNull()) {
+                    return;
+                }
+                if (!lockedPointer->isLocked()) {
+                    lockedPointer.clear();
+                    disconnect(constrainsOneShotConnection);
+                    constrainsOneShotConnection = QMetaObject::Connection();
+                    disconnect(constrainsUnboundConnection);
+                    constrainsUnboundConnection = QMetaObject::Connection();
+                    emit q_func()->pointerConstraintsChanged();
+                }
+            }
+        );
+    }
+    constrainsUnboundConnection = QObject::connect(lock, &LockedPointerInterface::unbound, q_func(),
+        [this] {
+            if (lockedPointer.isNull()) {
+                return;
+            }
+            lockedPointer.clear();
+            disconnect(constrainsOneShotConnection);
+            constrainsOneShotConnection = QMetaObject::Connection();
+            disconnect(constrainsUnboundConnection);
+            constrainsUnboundConnection = QMetaObject::Connection();
+            emit q_func()->pointerConstraintsChanged();
+        }
+    );
+    emit q_func()->pointerConstraintsChanged();
+}
+
+void SurfaceInterface::Private::installPointerConstraint(ConfinedPointerInterface *confinement)
+{
+    Q_ASSERT(lockedPointer.isNull());
+    Q_ASSERT(confinedPointer.isNull());
+    confinedPointer = QPointer<ConfinedPointerInterface>(confinement);
+    if (confinement->lifeTime() == ConfinedPointerInterface::LifeTime::OneShot) {
+        constrainsOneShotConnection = QObject::connect(confinement, &ConfinedPointerInterface::confinedChanged, q_func(),
+            [this] {
+                if (confinedPointer.isNull()) {
+                    return;
+                }
+                if (!confinedPointer->isConfined()) {
+                    confinedPointer.clear();
+                    disconnect(constrainsOneShotConnection);
+                    constrainsOneShotConnection = QMetaObject::Connection();
+                    disconnect(constrainsUnboundConnection);
+                    constrainsUnboundConnection = QMetaObject::Connection();
+                    emit q_func()->pointerConstraintsChanged();
+                }
+            }
+        );
+    }
+    constrainsUnboundConnection = QObject::connect(confinement, &ConfinedPointerInterface::unbound, q_func(),
+        [this] {
+            if (confinedPointer.isNull()) {
+                return;
+            }
+            confinedPointer.clear();
+            disconnect(constrainsOneShotConnection);
+            constrainsOneShotConnection = QMetaObject::Connection();
+            disconnect(constrainsUnboundConnection);
+            constrainsUnboundConnection = QMetaObject::Connection();
+            emit q_func()->pointerConstraintsChanged();
+        }
+    );
+    emit q_func()->pointerConstraintsChanged();
+}
+
 #ifndef DOXYGEN_SHOULD_SKIP_THIS
 const struct wl_surface_interface SurfaceInterface::Private::s_interface = {
     resourceDestroyedCallback,
@@ -324,6 +401,12 @@ void SurfaceInterface::Private::swapStates(State *source, State *target, bool em
         target->transform = source->transform;
         target->transformIsSet = true;
     }
+    if (!lockedPointer.isNull()) {
+        lockedPointer->d_func()->commit();
+    }
+    if (!confinedPointer.isNull()) {
+        confinedPointer->d_func()->commit();
+    }
 
     *source = State{};
     source->children = target->children;
@@ -335,6 +418,9 @@ void SurfaceInterface::Private::swapStates(State *source, State *target, bool em
     }
     if (scaleFactorChanged) {
         emit q->scaleChanged(target->scale);
+        if (buffer && !sizeChanged) {
+            emit q->sizeChanged();
+        }
     }
     if (transformChanged) {
         emit q->transformChanged(target->transform);
@@ -634,9 +720,9 @@ QPointer< SubSurfaceInterface > SurfaceInterface::subSurface() const
 QSize SurfaceInterface::size() const
 {
     Q_D();
-    // TODO: apply transform and scale to the buffer size
+    // TODO: apply transform to the buffer size
     if (d->current.buffer) {
-        return d->current.buffer->size();
+        return d->current.buffer->size() / scale();
     }
     return QSize();
 }
@@ -707,19 +793,25 @@ void SurfaceInterface::setOutputs(const QVector<OutputInterface *> &outputs)
         for (wl_resource *r : resources) {
             wl_surface_send_leave(d->resource, r);
         }
+        disconnect(d->outputDestroyedConnections.take(*it));
     }
-    // TODO: send leave when OutputInterface gets destroyed
-
     QVector<OutputInterface *> addedOutputsOutputs = outputs;
     for (auto it = d->outputs.constBegin(), end = d->outputs.constEnd(); it != end; ++it) {
         const auto o = *it;
         addedOutputsOutputs.removeOne(o);
     }
     for (auto it = addedOutputsOutputs.constBegin(), end = addedOutputsOutputs.constEnd(); it != end; ++it) {
-        const auto resources = (*it)->clientResources(client());
+        const auto o = *it;
+        const auto resources = o->clientResources(client());
         for (wl_resource *r : resources) {
             wl_surface_send_enter(d->resource, r);
         }
+        d->outputDestroyedConnections[o] = connect(o, &Global::aboutToDestroyGlobal, this, [this, o] {
+            Q_D();
+            auto outputs = d->outputs;
+            if (outputs.removeOne(o)) {
+                setOutputs(outputs);
+            }});
     }
     // TODO: send enter when the client binds the OutputInterface another time
 
@@ -752,6 +844,18 @@ SurfaceInterface *SurfaceInterface::surfaceAt(const QPointF &position)
     return nullptr;
 }
 
+QPointer<LockedPointerInterface> SurfaceInterface::lockedPointer() const
+{
+    Q_D();
+    return d->lockedPointer;
+}
+
+QPointer<ConfinedPointerInterface> SurfaceInterface::confinedPointer() const
+{
+    Q_D();
+    return d->confinedPointer;
+}
+
 SurfaceInterface::Private *SurfaceInterface::d_func() const
 {
     return reinterpret_cast<Private*>(d.data());
